from dataclasses import dataclass
from typing import Any, Dict, Generator, List, Optional

try:
    import boto3  # type:ignore
    from botocore.client import ClientError
    from botocore.exceptions import ParamValidationError

    S3NativeClient = Any
    S3NativeBucket = Any
except (ImportError, ModuleNotFoundError):
    raise ImportError(
        """You are using the S3 functionality of Pathy without
    having the required dependencies installed.

    Please try installing them:

        pip install pathy[s3]

    """
    )

from . import (
    Blob,
    Bucket,
    BucketClient,
    BucketEntry,
    PathyScanDir,
    PurePathy,
    register_client,
)


class BucketEntryS3(BucketEntry):
    bucket: "BucketS3"
    raw: Any


@dataclass
class BlobS3(Blob):
    client: S3NativeClient
    bucket: "BucketS3"

    def delete(self) -> None:
        self.client.delete_object(Bucket=self.bucket.name, Key=self.name)

    def exists(self) -> bool:
        response = self.client.list_objects_v2(
            Bucket=self.bucket.name, Prefix=self.name
        )
        objects = response.get("Contents", [])
        matched = [o["Key"] for o in objects if o["Key"] == self.name]
        return len(matched) > 0


@dataclass
class BucketS3(Bucket):
    name: str
    client: S3NativeClient
    bucket: S3NativeBucket

    def get_blob(self, blob_name: str) -> Optional[BlobS3]:
        blob_stat: Dict[str, Any]
        try:
            blob_stat = self.client.head_object(Bucket=self.name, Key=blob_name)
        except ClientError:
            return None
        updated = blob_stat["LastModified"].timestamp()
        size = blob_stat["ContentLength"]
        return BlobS3(
            client=self.client,
            bucket=self,
            owner=None,  # type:ignore
            name=blob_name,  # type:ignore
            raw=None,
            size=size,
            updated=int(updated),  # type:ignore
        )

    def copy_blob(  # type:ignore[override]
        self, blob: BlobS3, target: "BucketS3", name: str
    ) -> Optional[BlobS3]:
        source = {"Bucket": blob.bucket.name, "Key": blob.name}
        self.client.copy(source, target.name, name)
        pathy_blob: Optional[BlobS3] = self.get_blob(name)
        assert pathy_blob is not None, "copy failed"
        assert pathy_blob.updated is not None, "new blobl has invalid updated time"
        return BlobS3(
            client=self.client,
            bucket=self.bucket,
            owner=None,
            name=name,
            raw=pathy_blob,
            size=pathy_blob.size,
            updated=pathy_blob.updated,
        )

    def delete_blob(self, blob: BlobS3) -> None:  # type:ignore[override]
        self.client.delete_object(Bucket=self.name, Key=blob.name)

    def delete_blobs(self, blobs: List[BlobS3]) -> None:  # type:ignore[override]
        for blob in blobs:
            self.delete_blob(blob)

    def exists(self) -> bool:
        # TODO: are you sure this always holds?
        #
        # S3 buckets don't make it this far if they don't exist. The BucketS3 instance
        # is not instantiated unless a metadata check on the bucket passes.
        return True


class BucketClientS3(BucketClient):
    client: S3NativeClient
    _session: Optional[boto3.Session]

    @property
    def client_params(self) -> Dict[str, Any]:
        session: Any = self._session
        result: Any = dict() if session is None else dict(client=session.client("s3"))
        return result

    def __init__(self, **kwargs: Any) -> None:
        self.recreate(**kwargs)

    def recreate(self, **kwargs: Any) -> None:
        key_id = kwargs.get("key_id", None)
        key_secret = kwargs.get("key_secret", None)
        boto_session: Any = boto3
        if key_id is not None and key_secret is not None:
            self._session = boto_session = boto3.Session(  # type:ignore
                aws_access_key_id=key_id,
                aws_secret_access_key=key_secret,
            )
        self.client = boto_session.client("s3")  # type:ignore

    def make_uri(self, path: PurePathy) -> str:
        return str(path)

    def create_bucket(  # type:ignore[override]
        self, path: PurePathy
    ) -> S3NativeBucket:
        return self.client.create_bucket(Bucket=path.root)  # type:ignore

    def delete_bucket(self, path: PurePathy) -> None:
        self.client.delete_bucket(Bucket=path.root)

    def exists(self, path: PurePathy) -> bool:
        # Because we want all the parents of a valid blob (e.g. "directory" in
        # "directory/foo.file") to return True, we enumerate the blobs with a prefix
        # and compare the object names to see if they match a substring of the path
        key_name = str(path.key)
        for obj in self.list_blobs(path):
            if obj.name.startswith(key_name + path._flavour.sep):  # type:ignore
                return True
        return False

    def lookup_bucket(self, path: PurePathy) -> Optional[BucketS3]:
        try:
            return self.get_bucket(path)
        except FileNotFoundError:
            return None

    def get_bucket(self, path: PurePathy) -> BucketS3:
        try:
            native_bucket = self.client.head_bucket(Bucket=path.root)
            return BucketS3(str(path.root), client=self.client, bucket=native_bucket)
        except (ClientError, ParamValidationError):
            raise FileNotFoundError(f"Bucket {path.root} does not exist!")

    def list_buckets(  # type:ignore[override]
        self, **kwargs: Dict[str, Any]
    ) -> Generator[S3NativeBucket, None, None]:
        native_buckets = self.client.list_buckets(**kwargs)["Buckets"]
        results = (BucketS3(n["Name"], self.client, n) for n in native_buckets)
        return results

    def scandir(  # type:ignore[override]
        self,
        path: Optional[PurePathy] = None,
        prefix: Optional[str] = None,
        delimiter: Optional[str] = None,
    ) -> PathyScanDir:
        return ScanDirS3(client=self, path=path, prefix=prefix, delimiter=delimiter)

    def list_blobs(
        self,
        path: PurePathy,
        prefix: Optional[str] = None,
        delimiter: Optional[str] = None,
    ) -> Generator[BlobS3, None, None]:
        bucket = self.lookup_bucket(path)
        if bucket is None:
            return
        paginator = self.client.get_paginator("list_objects_v2")
        kwargs = {"Bucket": bucket.name}
        if prefix is not None:
            kwargs["Prefix"] = prefix
        for page in paginator.paginate(**kwargs):
            for item in page.get("Contents", []):
                yield BlobS3(
                    client=self.client,
                    bucket=bucket,
                    owner=None,
                    name=item["Key"],
                    raw=item,
                    size=item["Size"],
                    updated=int(item["LastModified"].timestamp()),
                )


class ScanDirS3(PathyScanDir):
    _client: BucketClientS3

    def __init__(
        self,
        client: BucketClient,
        path: Optional[PurePathy] = None,
        prefix: Optional[str] = None,
        delimiter: Optional[str] = None,
        page_size: Optional[int] = None,
    ) -> None:
        super().__init__(client=client, path=path, prefix=prefix, delimiter=delimiter)
        self._page_size = page_size

    def scandir(self) -> Generator[BucketEntryS3, None, None]:
        if self._path is None or not self._path.root:
            s3_bucket: BucketS3
            for s3_bucket in self._client.list_buckets():
                yield BucketEntryS3(s3_bucket.name, is_dir=True, raw=None)
            return
        sep = self._path._flavour.sep  # type:ignore
        bucket = self._client.lookup_bucket(self._path)
        if bucket is None:
            return

        kwargs: Any = {"Bucket": bucket.name, "Delimiter": sep}
        if self._prefix is not None:
            kwargs["Prefix"] = self._prefix
        if self._page_size is not None:
            kwargs["MaxKeys"] = self._page_size
        continuation_token: Optional[str] = None
        while True:
            if continuation_token:
                kwargs["ContinuationToken"] = continuation_token
            response = self._client.client.list_objects_v2(**kwargs)
            for folder in response.get("CommonPrefixes", []):
                prefix = folder["Prefix"]
                full_name = prefix[:-1] if prefix.endswith(sep) else prefix
                name = full_name.split(sep)[-1]
                yield BucketEntryS3(name, is_dir=True)
            for file in response.get("Contents", ()):
                name = file["Key"].split(sep)[-1]
                yield BucketEntryS3(
                    name=name,
                    is_dir=False,
                    size=file["Size"],
                    last_modified=int(file["LastModified"].timestamp()),
                )
            if not response.get("IsTruncated"):
                break
            continuation_token = response.get("NextContinuationToken")


register_client("s3", BucketClientS3)



has_gcs: bool
try:
    from ..gcs import BucketClientGCS

    has_gcs = bool(BucketClientGCS)
except ImportError:
    has_gcs = False

has_s3: bool
try:
    from ..s3 import BucketClientS3

    has_s3 = bool(BucketClientS3)
except ImportError:
    has_s3 = False



import os
import sys
from pathlib import Path, PurePosixPath, PureWindowsPath

import pytest

from .. import PurePathy


def test_pure_pathy_paths_of_a_different_flavour() -> None:
    with pytest.raises(TypeError):
        PurePathy("/bucket/key") < PurePosixPath("/bucket/key")

    with pytest.raises(TypeError):
        PureWindowsPath("/bucket/key") > PurePathy("/bucket/key")


def test_pure_pathy_repr() -> None:
    a = PurePathy("/var/tests/fake")
    assert a.as_posix() == "/var/tests/fake"
    assert repr(PurePathy("fake_file.txt")) == "PurePathy('fake_file.txt')"
    assert str(PurePathy("fake_file.txt")) == "fake_file.txt"
    assert bytes(PurePathy("fake_file.txt")) == b"fake_file.txt"


def test_pure_pathy_scheme_extraction() -> None:
    assert PurePathy("gs://var/tests/fake").scheme == "gs"
    assert PurePathy("s3://var/tests/fake").scheme == "s3"
    assert PurePathy("file://var/tests/fake").scheme == "file"
    assert PurePathy("/var/tests/fake").scheme == ""


@pytest.mark.skipif(sys.version_info < (3, 6), reason="requires python3.6 or higher")
def test_pure_pathy_fspath() -> None:
    assert os.fspath(PurePathy("/var/tests/fake")) == "/var/tests/fake"


def test_pure_pathy_join_strs() -> None:
    assert PurePathy("foo", "some/path", "bar") == PurePathy("foo/some/path/bar")


def test_pure_pathy_parse_parts() -> None:
    # Needs two parts to extract scheme/bucket
    with pytest.raises(ValueError):
        PurePathy("foo:")

    assert PurePathy("foo://bar") is not None


def test_pure_pathy_join_paths() -> None:
    assert PurePathy(Path("foo"), Path("bar")) == PurePathy("foo/bar")


def test_pure_pathy_empty() -> None:
    assert PurePathy() == PurePathy(".")


def test_pure_pathy_absolute_paths() -> None:
    assert PurePathy("/etc", "/usr", "lib64") == PurePathy("/usr/lib64")


def test_pure_pathy_slashes_single_double_dots() -> None:
    assert PurePathy("foo//bar") == PurePathy("foo/bar")
    assert PurePathy("foo/./bar") == PurePathy("foo/bar")
    assert PurePathy("foo/../bar") == PurePathy("bar")
    assert PurePathy("../bar") == PurePathy("../bar")
    assert PurePathy("foo", "../bar") == PurePathy("bar")


def test_pure_pathy_operators() -> None:
    assert PurePathy("/etc") / "init.d" / "apache2" == PurePathy("/etc/init.d/apache2")
    assert "/var" / PurePathy("tests") / "fake" == PurePathy("/var/tests/fake")


def test_pure_pathy_parts() -> None:
    assert PurePathy("../bar").parts == ("..", "bar")
    assert PurePathy("foo//bar").parts == ("foo", "bar")
    assert PurePathy("foo/./bar").parts == ("foo", "bar")
    assert PurePathy("foo/../bar").parts == ("bar",)
    assert PurePathy("foo", "../bar").parts == ("bar",)
    assert PurePathy("/foo/bar").parts == ("/", "foo", "bar")


def test_pure_pathy_prefix() -> None:
    assert PurePathy("gs://bar").prefix == ""
    assert PurePathy("gs://bar/baz/").prefix == "baz/"
    assert PurePathy("gs://bar/baz").prefix == "baz/"


def test_pure_pathy_drive() -> None:
    assert PurePathy("foo//bar").drive == ""
    assert PurePathy("foo/./bar").drive == ""
    assert PurePathy("foo/../bar").drive == ""
    assert PurePathy("../bar").drive == ""
    assert PurePathy("foo", "../bar").drive == ""
    assert PurePathy("/foo/bar").drive == ""


def test_pure_pathy_root() -> None:
    assert PurePathy("foo//bar").root == ""
    assert PurePathy("foo/./bar").root == ""
    assert PurePathy("foo/../bar").root == ""
    assert PurePathy("../bar").root == ""
    assert PurePathy("foo", "../bar").root == ""
    assert PurePathy("/foo/bar").root == "/"


def test_pure_pathy_anchor() -> None:
    assert PurePathy("foo//bar").anchor == ""
    assert PurePathy("foo/./bar").anchor == ""
    assert PurePathy("foo/../bar").anchor == ""
    assert PurePathy("../bar").anchor == ""
    assert PurePathy("foo", "../bar").anchor == ""
    assert PurePathy("/foo/bar").anchor == "/"


def test_pure_pathy_parents() -> None:
    assert tuple(PurePathy("foo//bar").parents) == (
        PurePathy("foo"),
        PurePathy("."),
    )
    assert tuple(PurePathy("foo/./bar").parents) == (
        PurePathy("foo"),
        PurePathy("."),
    )
    assert tuple(PurePathy("foo/../bar").parents) == (PurePathy("."),)
    assert tuple(PurePathy("../bar").parents) == (PurePathy(".."), PurePathy("."))
    assert tuple(PurePathy("foo", "../bar").parents) == (PurePathy("."),)
    assert tuple(PurePathy("/foo/bar").parents) == (
        PurePathy("/foo"),
        PurePathy("/"),
    )


def test_pure_pathy_parent() -> None:
    assert PurePathy("foo//bar").parent == PurePathy("foo")
    assert PurePathy("foo/./bar").parent == PurePathy("foo")
    assert PurePathy("foo/../bar").parent == PurePathy(".")
    assert PurePathy("../bar").parent == PurePathy("..")
    assert PurePathy("foo", "../bar").parent == PurePathy(".")
    assert PurePathy("/foo/bar").parent == PurePathy("/foo")
    assert PurePathy(".").parent == PurePathy(".")
    assert PurePathy("/").parent == PurePathy("/")


def test_pure_pathy_name() -> None:
    assert PurePathy("my/library/fake_file.txt").name == "fake_file.txt"


def test_pure_pathy_suffix() -> None:
    assert PurePathy("my/library/fake_file.txt").suffix == ".txt"
    assert PurePathy("my/library.tar.gz").suffix == ".gz"
    assert PurePathy("my/library").suffix == ""


def test_pure_pathy_suffixes() -> None:
    assert PurePathy("my/library.tar.gar").suffixes == [".tar", ".gar"]
    assert PurePathy("my/library.tar.gz").suffixes == [".tar", ".gz"]
    assert PurePathy("my/library").suffixes == []


def test_pure_pathy_stem() -> None:
    assert PurePathy("my/library.tar.gar").stem == "library.tar"
    assert PurePathy("my/library.tar").stem == "library"
    assert PurePathy("my/library").stem == "library"


def test_pure_pathy_uri() -> None:
    assert PurePathy("/etc/passwd").as_uri() == "/etc/passwd"
    assert PurePathy("/etc/init.d/apache2").as_uri() == "/etc/init.d/apache2"
    assert PurePathy("/bucket/key").as_uri() == "/bucket/key"


def test_pure_pathy_absolute() -> None:
    assert PurePathy("/a/b").is_absolute()
    assert not PurePathy("a/b").is_absolute()


def test_pure_pathy_reserved() -> None:
    assert not PurePathy("/a/b").is_reserved()
    assert not PurePathy("a/b").is_reserved()


def test_pure_pathy_joinpath() -> None:
    assert PurePathy("/etc").joinpath("passwd") == PurePathy("/etc/passwd")
    assert PurePathy("/etc").joinpath(PurePathy("passwd")) == PurePathy("/etc/passwd")
    assert PurePathy("/etc").joinpath("init.d", "apache2") == PurePathy(
        "/etc/init.d/apache2"
    )


def test_pure_pathy_match() -> None:
    assert PurePathy("a/b.py").match("*.py")
    assert PurePathy("/a/b/c.py").match("b/*.py")
    assert not PurePathy("/a/b/c.py").match("a/*.py")
    assert PurePathy("/a.py").match("/*.py")
    assert not PurePathy("a/b.py").match("/*.py")
    assert not PurePathy("a/b.py").match("*.Py")


def test_pure_pathy_relative_to() -> None:
    gcs_path = PurePathy("/etc/passwd")
    assert gcs_path.relative_to("/") == PurePathy("etc/passwd")
    assert gcs_path.relative_to("/etc") == PurePathy("passwd")
    with pytest.raises(ValueError):
        gcs_path.relative_to("/usr")


def test_pure_pathy_with_name() -> None:
    gcs_path = PurePathy("/Downloads/pathlib.tar.gz")
    assert gcs_path.with_name("fake_file.txt") == PurePathy("/Downloads/fake_file.txt")
    gcs_path = PurePathy("/")
    with pytest.raises(ValueError):
        gcs_path.with_name("fake_file.txt")


def test_pure_pathy_with_suffix() -> None:
    gcs_path = PurePathy("/Downloads/pathlib.tar.gz")
    assert gcs_path.with_suffix(".bz2") == PurePathy("/Downloads/pathlib.tar.bz2")
    gcs_path = PurePathy("README")
    assert gcs_path.with_suffix(".txt") == PurePathy("README.txt")
    gcs_path = PurePathy("README.txt")
    assert gcs_path.with_suffix("") == PurePathy("README")



import pathlib
import tempfile

import pytest
from typer.testing import CliRunner

from pathy import Pathy
from pathy.cli import app

from .conftest import ENV_ID, TEST_ADAPTERS

runner = CliRunner()

# TODO: add support/test for wildcard cp/mv/rm/ls paths (e.g. "pathy cp gs://my-bucket/*.file ./")
# TODO: add support/test for streaming in/out sources (e.g. "pathy cp - gs://my-bucket/my.file")


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_cp_invalid_from_path(with_adapter: str, bucket: str) -> None:
    source = f"{with_adapter}://{bucket}/{ENV_ID}/cli_cp_file_invalid/file.txt"
    destination = f"{with_adapter}://{bucket}/{ENV_ID}/cli_cp_file_invalid/dest.txt"
    assert runner.invoke(app, ["cp", source, destination]).exit_code == 1
    assert not Pathy(destination).is_file()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_cp_file(with_adapter: str, bucket: str) -> None:
    source = f"{with_adapter}://{bucket}/{ENV_ID}/cli_cp_file/file.txt"
    destination = f"{with_adapter}://{bucket}/{ENV_ID}/cli_cp_file/other.txt"
    Pathy(source).write_text("---")
    assert runner.invoke(app, ["cp", source, destination]).exit_code == 0
    assert Pathy(source).exists()
    assert Pathy(destination).is_file()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_cp_file_name_from_source(with_adapter: str, bucket: str) -> None:
    source = pathlib.Path("./file.txt")
    source.touch()
    destination = f"{with_adapter}://{bucket}/{ENV_ID}/cli_cp_file/"
    assert runner.invoke(app, ["cp", str(source), destination]).exit_code == 0
    assert Pathy(f"{destination}file.txt").is_file()
    source.unlink()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_cp_folder(with_adapter: str, bucket: str) -> None:
    root = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/")
    source = root / "cli_cp_folder"
    destination = root / "cli_cp_folder_other"
    for i in range(2):
        for j in range(2):
            (source / f"{i}" / f"{j}").write_text("---")
    assert runner.invoke(app, ["cp", str(source), str(destination)]).exit_code == 0
    assert Pathy(source).exists()
    assert Pathy(destination).is_dir()
    for i in range(2):
        for j in range(2):
            assert (destination / f"{i}" / f"{j}").is_file()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_mv_folder(with_adapter: str, bucket: str) -> None:
    root = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/")
    source = root / "cli_mv_folder"
    destination = root / "cli_mv_folder_other"
    for i in range(2):
        for j in range(2):
            (source / f"{i}" / f"{j}").write_text("---")
    assert runner.invoke(app, ["mv", str(source), str(destination)]).exit_code == 0
    assert not Pathy(source).exists()
    assert Pathy(destination).is_dir()
    # Ensure source files are gone
    for i in range(2):
        for j in range(2):
            assert not (source / f"{i}" / f"{j}").is_file()
    # And dest files exist
    for i in range(2):
        for j in range(2):
            assert (destination / f"{i}" / f"{j}").is_file()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_mv_file_copy_from_name(with_adapter: str, bucket: str) -> None:
    source = pathlib.Path("./file.txt")
    source.touch()
    destination = f"{with_adapter}://{bucket}/{ENV_ID}/cli_cp_file/"
    assert runner.invoke(app, ["mv", str(source), destination]).exit_code == 0
    assert Pathy(f"{destination}file.txt").is_file()
    # unlink should happen from the operation
    assert not source.exists()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_mv_file(with_adapter: str, bucket: str) -> None:
    source = f"{with_adapter}://{bucket}/{ENV_ID}/cli_mv_file/file.txt"
    destination = f"{with_adapter}://{bucket}/{ENV_ID}/cli_mv_file/other.txt"
    Pathy(source).write_text("---")
    assert Pathy(source).exists()
    assert runner.invoke(app, ["mv", source, destination]).exit_code == 0
    assert not Pathy(source).exists()
    assert Pathy(destination).is_file()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_mv_file_across_buckets(
    with_adapter: str, bucket: str, other_bucket: str
) -> None:
    source = f"{with_adapter}://{bucket}/{ENV_ID}/cli_mv_file_across_buckets/file.txt"
    destination = (
        f"{with_adapter}://{other_bucket}/{ENV_ID}/cli_mv_file_across_buckets/other.txt"
    )
    Pathy(source).write_text("---")
    assert Pathy(source).exists()
    assert runner.invoke(app, ["mv", source, destination]).exit_code == 0
    assert not Pathy(source).exists()
    assert Pathy(destination).is_file()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_mv_folder_across_buckets(
    with_adapter: str, bucket: str, other_bucket: str
) -> None:
    source = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/cli_mv_folder_across_buckets")
    destination = Pathy(
        f"{with_adapter}://{other_bucket}/{ENV_ID}/cli_mv_folder_across_buckets"
    )
    for i in range(2):
        for j in range(2):
            (source / f"{i}" / f"{j}").write_text("---")
    assert runner.invoke(app, ["mv", str(source), str(destination)]).exit_code == 0
    assert not Pathy(source).exists()
    assert Pathy(destination).is_dir()
    # Ensure source files are gone
    for i in range(2):
        for j in range(2):
            assert not (source / f"{i}" / f"{j}").is_file()
    # And dest files exist
    for i in range(2):
        for j in range(2):
            assert (destination / f"{i}" / f"{j}").is_file()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_rm_invalid_file(with_adapter: str, bucket: str) -> None:
    source = f"{with_adapter}://{bucket}/{ENV_ID}/cli_rm_file_invalid/file.txt"
    path = Pathy(source)
    assert not path.exists()
    assert runner.invoke(app, ["rm", source]).exit_code == 1


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_rm_file(with_adapter: str, bucket: str) -> None:
    source = f"{with_adapter}://{bucket}/{ENV_ID}/cli_rm_file/file.txt"
    path = Pathy(source)
    path.write_text("---")
    assert path.exists()
    assert runner.invoke(app, ["rm", source]).exit_code == 0
    assert not path.exists()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_rm_verbose(with_adapter: str, bucket: str) -> None:
    root = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/") / "cli_rm_folder"
    source = str(root / "file.txt")
    other = str(root / "folder/other")
    Pathy(source).write_text("---")
    Pathy(other).write_text("---")
    result = runner.invoke(app, ["rm", "-v", source])
    assert result.exit_code == 0
    assert source in result.output
    assert other not in result.output

    Pathy(source).write_text("---")
    result = runner.invoke(app, ["rm", "-rv", str(root)])
    assert result.exit_code == 0
    assert source in result.output
    assert other in result.output


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_rm_folder(with_adapter: str, bucket: str) -> None:
    root = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/")
    source = root / "cli_rm_folder"
    for i in range(2):
        for j in range(2):
            (source / f"{i}" / f"{j}").write_text("---")

    # Returns exit code 1 without recursive flag when given a folder
    assert runner.invoke(app, ["rm", str(source)]).exit_code == 1
    assert runner.invoke(app, ["rm", "-r", str(source)]).exit_code == 0
    assert not Pathy(source).exists()
    # Ensure source files are gone
    for i in range(2):
        for j in range(2):
            assert not (source / f"{i}" / f"{j}").is_file()


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_ls_invalid_source(with_adapter: str, bucket: str) -> None:
    root = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/") / "cli_ls_invalid"
    three = str(root / "folder/file.txt")

    result = runner.invoke(app, ["ls", str(three)])
    assert result.exit_code == 1
    assert "No such file or directory" in result.output


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_ls(with_adapter: str, bucket: str) -> None:
    root = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/") / "cli_ls"
    one = str(root / "file.txt")
    two = str(root / "other.txt")
    three = str(root / "folder/file.txt")
    Pathy(one).write_text("---")
    Pathy(two).write_text("---")
    Pathy(three).write_text("---")

    result = runner.invoke(app, ["ls", str(root)])
    assert result.exit_code == 0
    assert one in result.output
    assert two in result.output
    assert str(root / "folder") in result.output

    result = runner.invoke(app, ["ls", "-l", str(root)])
    assert result.exit_code == 0
    assert one in result.output
    assert two in result.output
    assert str(root / "folder") in result.output


@pytest.mark.parametrize("adapter", TEST_ADAPTERS)
def test_cli_ls_local_files(with_adapter: str, bucket: str) -> None:
    root = Pathy.fluid(tempfile.mkdtemp()) / ENV_ID / "ls"
    root.mkdir(parents=True, exist_ok=True)
    for i in range(3):
        (root / f"file_{i}").write_text("NICE")
    files = list(root.ls())
    assert len(files) == 3
    valid_names = [f"file_{i}" for i in range(len(files))]
    for i, blob_stat in enumerate(files):
        assert blob_stat.name in valid_names
        assert blob_stat.size == 4
        assert blob_stat.last_modified is not None

    root = Pathy(f"{with_adapter}://{bucket}/{ENV_ID}/") / "cli_ls"
    one = str(root / "file.txt")
    two = str(root / "other.txt")
    three = str(root / "folder/file.txt")
    Pathy(one).write_text("---")
    Pathy(two).write_text("---")
    Pathy(three).write_text("---")

    result = runner.invoke(app, ["ls", str(root)])
    assert result.exit_code == 0
    assert one in result.output
    assert two in result.output
    assert str(root / "folder") in result.output

    result = runner.invoke(app, ["ls", "-l", str(root)])
    assert result.exit_code == 0
    assert one in result.output
    assert two in result.output
    assert str(root / "folder") in result.output



from dataclasses import dataclass
from typing import Any, Dict, Generator, List, Optional

try:
    from google.api_core.exceptions import BadRequest  # type:ignore
    from google.cloud.storage import Blob as GCSNativeBlob  # type:ignore
    from google.cloud.storage import Bucket as GCSNativeBucket  # type:ignore
    from google.cloud.storage import Client as GCSNativeClient  # type:ignore
except (ImportError, ModuleNotFoundError):
    raise ImportError(
        """You are using the GCS functionality of Pathy without
    having the required dependencies installed.

    Please try installing them:

        pip install pathy[gcs]

    """
    )

from . import (
    Blob,
    Bucket,
    BucketClient,
    BucketEntry,
    PathyScanDir,
    PurePathy,
    register_client,
)


class BucketEntryGCS(BucketEntry):
    bucket: "BucketGCS"
    raw: GCSNativeBlob  # type:ignore[override]


@dataclass
class BlobGCS(Blob):
    def delete(self) -> None:
        self.raw.delete()  # type:ignore

    def exists(self) -> bool:
        return self.raw.exists()  # type:ignore


@dataclass
class BucketGCS(Bucket):
    name: str
    bucket: GCSNativeBucket

    def get_blob(self, blob_name: str) -> Optional[BlobGCS]:
        assert isinstance(
            blob_name, str
        ), f"expected str blob name, but found: {type(blob_name)}"
        native_blob: Optional[Any] = self.bucket.get_blob(blob_name)  # type:ignore
        if native_blob is None:
            return None
        return BlobGCS(
            bucket=self.bucket,
            owner=native_blob.owner,  # type:ignore
            name=native_blob.name,  # type:ignore
            raw=native_blob,
            size=native_blob.size,
            updated=int(native_blob.updated.timestamp()),  # type:ignore
        )

    def copy_blob(  # type:ignore[override]
        self, blob: BlobGCS, target: "BucketGCS", name: str
    ) -> Optional[BlobGCS]:
        assert blob.raw is not None, "raw storage.Blob instance required"
        native_blob: GCSNativeBlob = self.bucket.copy_blob(  # type: ignore
            blob.raw, target.bucket, name
        )
        return BlobGCS(
            bucket=self.bucket,
            owner=native_blob.owner,  # type:ignore
            name=native_blob.name,  # type:ignore
            raw=native_blob,
            size=native_blob.size,
            updated=int(native_blob.updated.timestamp()),  # type:ignore
        )

    def delete_blob(self, blob: BlobGCS) -> None:  # type:ignore[override]
        return self.bucket.delete_blob(blob.name)  # type:ignore

    def delete_blobs(self, blobs: List[BlobGCS]) -> None:  # type:ignore[override]
        return self.bucket.delete_blobs(blobs)  # type:ignore

    def exists(self) -> bool:
        return self.bucket.exists()  # type:ignore


class BucketClientGCS(BucketClient):
    client: GCSNativeClient

    @property
    def client_params(self) -> Any:
        return dict(client=self.client)

    def __init__(self, **kwargs: Any) -> None:
        self.recreate(**kwargs)

    def recreate(self, **kwargs: Any) -> None:
        creds = kwargs["credentials"] if "credentials" in kwargs else None
        if creds is not None:
            kwargs["project"] = creds.project_id
        self.client = GCSNativeClient(**kwargs)

    def make_uri(self, path: PurePathy) -> str:
        return str(path)

    def create_bucket(  # type:ignore[override]
        self, path: PurePathy
    ) -> GCSNativeBucket:
        return self.client.create_bucket(path.root)  # type:ignore

    def delete_bucket(self, path: PurePathy) -> None:
        bucket = self.client.get_bucket(path.root)  # type:ignore
        bucket.delete()  # type:ignore

    def exists(self, path: PurePathy) -> bool:
        # Because we want all the parents of a valid blob (e.g. "directory" in
        # "directory/foo.file") to return True, we enumerate the blobs with a prefix
        # and compare the object names to see if they match a substring of the path
        key_name = str(path.key)
        for obj in self.list_blobs(path):
            if obj.name.startswith(key_name + path._flavour.sep):  # type:ignore
                return True
        return False

    def lookup_bucket(self, path: PurePathy) -> Optional[BucketGCS]:
        try:
            return self.get_bucket(path)
        except FileNotFoundError:
            return None

    def get_bucket(self, path: PurePathy) -> BucketGCS:
        native_bucket: Any = self.client.bucket(path.root)  # type:ignore
        try:
            if native_bucket.exists():
                return BucketGCS(str(path.root), bucket=native_bucket)
        except BadRequest:
            pass
        raise FileNotFoundError(f"Bucket {path.root} does not exist!")

    def list_buckets(  # type:ignore[override]
        self, **kwargs: Dict[str, Any]
    ) -> Generator[GCSNativeBucket, None, None]:
        return self.client.list_buckets(**kwargs)  # type:ignore

    def scandir(  # type:ignore[override]
        self,
        path: Optional[PurePathy] = None,
        prefix: Optional[str] = None,
        delimiter: Optional[str] = None,
    ) -> PathyScanDir:
        return ScanDirGCS(client=self, path=path, prefix=prefix, delimiter=delimiter)

    def list_blobs(
        self,
        path: PurePathy,
        prefix: Optional[str] = None,
        delimiter: Optional[str] = None,
    ) -> Generator[BlobGCS, None, None]:
        bucket = self.lookup_bucket(path)
        if bucket is None:
            return
        response: Any = self.client.list_blobs(  # type:ignore
            path.root, prefix=prefix, delimiter=delimiter
        )
        for page in response.pages:  # type:ignore
            for item in page:
                yield BlobGCS(
                    bucket=bucket,
                    owner=item.owner,
                    name=item.name,
                    raw=item,
                    size=item.size,
                    updated=item.updated.timestamp(),
                )


class ScanDirGCS(PathyScanDir):
    _client: BucketClientGCS

    def scandir(self) -> Generator[BucketEntryGCS, None, None]:
        if self._path is None or not self._path.root:
            gcs_bucket: GCSNativeBucket
            for gcs_bucket in self._client.list_buckets():
                yield BucketEntryGCS(
                    gcs_bucket.name, is_dir=True, raw=None  # type:ignore
                )
            return
        sep = self._path._flavour.sep  # type:ignore
        bucket = self._client.lookup_bucket(self._path)
        if bucket is None:
            return
        response = self._client.client.list_blobs(  # type:ignore
            bucket.name, prefix=self._prefix, delimiter=sep
        )
        for page in response.pages:  # type:ignore
            folder: str
            for folder in list(page.prefixes):  # type:ignore
                full_name = folder[:-1] if folder.endswith(sep) else folder
                name = full_name.split(sep)[-1]
                if name:
                    yield BucketEntryGCS(name, is_dir=True, raw=None)
            item: Any
            for item in page:  # type:ignore
                name = item.name.split(sep)[-1]
                if name:
                    yield BucketEntryGCS(
                        name=name,
                        is_dir=False,
                        size=item.size,
                        last_modified=item.updated.timestamp(),
                        raw=item,
                    )


register_client("gs", BucketClientGCS)



from unittest.mock import patch

import pytest

from pathy import Pathy, get_client, set_client_params, use_fs

from . import has_gcs

GCS_ADAPTER = ["gcs"]


def raise_default_creds_error() -> None:
    from google.auth.exceptions import DefaultCredentialsError

    raise DefaultCredentialsError()


@pytest.mark.parametrize("adapter", GCS_ADAPTER)
@patch("google.auth.default", raise_default_creds_error)
@pytest.mark.skipif(not has_gcs, reason="requires gcs")
def test_gcs_default_credentials_error_is_preserved(
    with_adapter: str, bucket: str
) -> None:
    from google.auth.exceptions import DefaultCredentialsError

    # Remove the default credentials (this will recreate the client and error)
    with pytest.raises(DefaultCredentialsError):
        set_client_params("gs")


@pytest.mark.parametrize("adapter", GCS_ADAPTER)
@pytest.mark.skipif(not has_gcs, reason="requires gcs")
def test_gcs_as_uri(with_adapter: str, bucket: str) -> None:
    assert Pathy("gs://etc/passwd").as_uri() == "gs://etc/passwd"
    assert Pathy("gs://etc/init.d/apache2").as_uri() == "gs://etc/init.d/apache2"
    assert Pathy("gs://bucket/key").as_uri() == "gs://bucket/key"


@pytest.mark.skipif(has_gcs, reason="requires gcs deps to NOT be installed")
def test_gcs_import_error_missing_deps() -> None:
    use_fs(False)
    with pytest.raises(ImportError):
        get_client("gs")


@pytest.mark.parametrize("adapter", GCS_ADAPTER)
@pytest.mark.skipif(not has_gcs, reason="requires gcs")
def test_gcs_scandir_list_buckets(
    with_adapter: str, bucket: str, other_bucket: str
) -> None:
    from pathy.gcs import ScanDirGCS

    root = Pathy("gs://foo/bar")
    client = root._accessor.client(root)  # type:ignore
    scandir = ScanDirGCS(client=client, path=Pathy())
    assert sorted([s.name for s in scandir]) == sorted([bucket, other_bucket])


@pytest.mark.parametrize("adapter", GCS_ADAPTER)
@pytest.mark.skipif(not has_gcs, reason="requires gcs")
def test_gcs_scandir_invalid_bucket_name(with_adapter: str) -> None:
    from pathy.gcs import ScanDirGCS

    root = Pathy("gs://invalid_h3gE_ds5daEf_Sdf15487t2n4/bar")
    client = root._accessor.client(root)  # type:ignore
    scandir = ScanDirGCS(client=client, path=root)
    assert len(list(scandir)) == 0


@pytest.mark.parametrize("adapter", GCS_ADAPTER)
@pytest.mark.skipif(not has_gcs, reason="requires gcs")
def test_gcs_bucket_client_list_blobs(with_adapter: str, bucket: str) -> None:
    """Test corner-case in GCS client that isn't easily reachable from Pathy"""
    from pathy.gcs import BucketClientGCS

    client: BucketClientGCS = get_client("gs")
    root = Pathy("gs://invalid_h3gE_ds5daEf_Sdf15487t2n4")
    assert len(list(client.list_blobs(root))) == 0
